// ps2 controller routines

#include <c8051f300.h>                       // SFR declarations
#include "ps2.h"

#define KEYB_RBUF_SIZE		16
#define MOUSE_RBUF_SIZE		64

extern uchar _shrp(uchar u);

// keyboard definitions
static char keyb_bitcount;
static uchar keyb_rbuf [KEYB_RBUF_SIZE];
uchar keyb_rin, keyb_rout;

// mouse definitions
static char mouse_bitcount;
static uchar mouse_rbuf [MOUSE_RBUF_SIZE];
uchar mouse_rin, mouse_rout;

sbit KDATA = P0^2;
sbit KCLK = P0^3;
sbit MDATA = P0^4;
sbit MCLK = P0^5;

void ps2_init(void)
{
	IT0 = 1;		// edge trigered INT0, INT1
	IT1 = 1;
	IT01CF = 0x53;	// active low, INT0 = P0.3, INT1 = P0.5
	EX0 = 1;		// enable external INT0, INT1
	EX1 = 1;

	// keyboard init
	keyb_bitcount = 11;
	keyb_rin = keyb_rout = 0;

	// mouse init
	mouse_bitcount = 11;
	mouse_rin = mouse_rout = 0;
}

#define SCANCODES_COUNT		136
#define E0_SCANCODES_COUNT	128

uchar code scan2_to_scan1[SCANCODES_COUNT] =
{
	0x00,0x43,0x00,0x3F,0x3D,0x3B,0x3C,0x58,
	0x00,0x44,0x42,0x40,0x3E,0x0F,0x29,0x00,
	0x00,0x38,0x2A,0x00,0x1D,0x10,0x02,0x00,
	0x00,0x00,0x2C,0x1F,0x1E,0x11,0x03,0x00,
	0x00,0x2E,0x2D,0x20,0x12,0x05,0x04,0x00,
	0x00,0x39,0x2F,0x21,0x14,0x13,0x06,0x00,
	0x00,0x31,0x30,0x23,0x22,0x15,0x07,0x00,
	0x00,0x00,0x32,0x24,0x16,0x08,0x09,0x00,
	0x00,0x33,0x25,0x17,0x18,0x0B,0x0A,0x00,
	0x00,0x34,0x35,0x26,0x27,0x19,0x0C,0x00,
	0x00,0x00,0x28,0x00,0x1A,0x0D,0x00,0x00,
	0x3A,0x36,0x1C,0x1B,0x00,0x2B,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x0E,0x00,
	0x00,0x4F,0x00,0x4B,0x47,0x00,0x00,0x00,
	0x52,0x53,0x50,0x4C,0x4D,0x48,0x01,0x45,
	0x57,0x4E,0x51,0x4A,0x37,0x49,0x46,0x00,
	0x00,0x00,0x00,0x41,0x00,0x00,0x00,0x00
};

uchar code e0_scan2_to_scan1[E0_SCANCODES_COUNT] =
{
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x65,0x38,0x2A,0x00,0x1D,0x10,0x00,0x00,
	0x66,0x00,0x00,0x00,0x00,0x00,0x00,0x5B,
	0x67,0x2E,0x00,0x20,0x00,0x00,0x00,0x5C,
	0x68,0x00,0x00,0x21,0x00,0x00,0x00,0x5D,
	0x69,0x00,0x30,0x00,0x22,0x00,0x00,0x5E,
	0x6A,0x00,0x32,0x24,0x00,0x00,0x00,0x5F,
	0x6B,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x6C,0x00,0x35,0x00,0x00,0x19,0x00,0x00,
	0x6D,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x1C,0x00,0x00,0x00,0x63,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x4F,0x00,0x4B,0x47,0x00,0x00,0x00,
	0x52,0x53,0x50,0x00,0x4D,0x48,0x00,0x00,
	0x00,0x00,0x51,0x00,0x37,0x49,0x00,0x00
};

uchar handle_at_scancode( uchar scancode)
{
	static uchar break_code;
	static uchar prev_scancode;

	switch( scancode)
	{
	case 0xf0:		// handle break code
		break_code = 0x80;
		return 0;

	case 0xe0:
	case 0xe1:		// escape sequence
		prev_scancode = scancode;
		break;

	case 0x00:
	case 0xff:
		prev_scancode = 0;
		break;

	default:
		if( prev_scancode)
		{
			switch( prev_scancode)
		 	{
			case 0xe1:		// handle e1 escape sequence
				{
					switch( scancode)
					{
					case 0x14:
						scancode = (0x1d | break_code);
						break;
					case 0x77:
						scancode = (0x45 | break_code);
					default:
						prev_scancode = 0;
					};
				} break;

			case 0xe0:		// handle e0 escape
				scancode = ( scancode < E0_SCANCODES_COUNT)? (e0_scan2_to_scan1[scancode] | break_code) : 0;

			default:
				prev_scancode = 0;
			};
		}
		else
		{
				scancode = ( scancode < SCANCODES_COUNT)? (scan2_to_scan1[scancode] | break_code) : 0;
	 	}
	};
	
   	break_code = 0;

  	return scancode;
}

static uchar kscan;	// holds keyboard scan code

// keyboard clock is connected to P0.3
// signal handler for external interrupt INT0
static void ps2_keyb_isr(void) interrupt 0
{
	if( keyb_bitcount > 0)
	{
	    if(--keyb_bitcount > 0)	// all bits received
	    {
			// bit 3 to 10 is data, parity bit, start and stop bits are ignored
	        if(keyb_bitcount < 10 && keyb_bitcount > 1)
	        {
	            kscan = (kscan >> 1);
	            if(KDATA)	// keyboard data is connected to P0.2
	                kscan = kscan | 0x80;
	        }
		}
		else
		{
			if (((keyb_rin - keyb_rout) & ~(KEYB_RBUF_SIZE-1)) == 0 && (kscan = handle_at_scancode(kscan)) != 0)
				keyb_rbuf [keyb_rin++ & (KEYB_RBUF_SIZE-1)] = kscan;
	
			keyb_bitcount = 11;
		}
	}
	else
	{
	    if( ++keyb_bitcount < 0)		// all bits received
	    {
			// 8 data bits and parity bit
	        if( keyb_bitcount < -1)
	        {
	            KDATA = (kscan & 0x01)? 1:0;
		        kscan = _shrp(kscan);		// shift right with parity
	        }
			else
			{
				// handle stop condition
				KDATA = 1;
			}
		}
		else
		{
			// setup rx mode
			KDATA = 1;
			keyb_bitcount = 11;
		}
	}			  
}

static uchar mscan;	// holds mouse scan code

// mouse clock is connected to P0.5
// signal handler for external interrupt INT1
static void ps2_mouse_isr(void) interrupt 2
{
	if( mouse_bitcount > 0)
	{
	    if(--mouse_bitcount > 0)	// all bits received
	    {
			// bit 3 to 10 is data, parity bit, start and stop bits are ignored
	        if(mouse_bitcount < 10 && mouse_bitcount > 1)
	        {
	            mscan = (mscan >> 1);
	            if(MDATA)	// mouse data is connected to P0.4
	                mscan = mscan | 0x80;
	        }
		}
		else
		{
			if (((mouse_rin - mouse_rout) & ~(MOUSE_RBUF_SIZE-1)) == 0)
				mouse_rbuf [mouse_rin++ & (MOUSE_RBUF_SIZE-1)] = mscan;
	
			mouse_bitcount = 11;
		}
	}
	else
	{
	    if( ++mouse_bitcount < 0)		// all bits received
	    {
			// 8 data bits and parity bit
	        if( mouse_bitcount < -1)
	        {
	            MDATA = (mscan & 0x01)? 1:0;
		        mscan = _shrp(mscan);		// shift right with parity
	        }
			else
			{
				// handle stop condition
				MDATA = 1;
			}
		}
		else
		{
			// setup rx mode
			MDATA = 1;
			mouse_bitcount = 11;
		}
	}
}

uchar ps2_keyb_getchar(void)
{
	return( keyb_rbuf [(keyb_rout++) & (KEYB_RBUF_SIZE - 1)]);
}

uchar ps2_mouse_getchar(void)
{
	return( mouse_rbuf [(mouse_rout++) & (MOUSE_RBUF_SIZE - 1)]);
}

static void udelay( unsigned int d)
{
	while( d-- > 0);
}

#define PS2_RTS_DELAY	500		// request to send(clk low) delay

void ps2_keyb_putchar( uchar u)
{
	// generate request to send
	KDATA = 1;
  	KCLK = 0;
	udelay(PS2_RTS_DELAY);		// wait > 100us (~250us)
	KDATA=0;
	udelay(100);		// 50us
  	KCLK=1;

	kscan = u;
	keyb_bitcount = -11;
}

void ps2_mouse_putchar( uchar u)
{
	// generate request to send
	MDATA = 1;
  	MCLK = 0;
	udelay(PS2_RTS_DELAY);		// wait > 100us (~250us)
	MDATA=0;
	udelay(100);		// 50us
  	MCLK=1;

	mscan = u;
	mouse_bitcount = -11;
}

